"""Game tiles."""

from dataclasses import dataclass
from enum import Enum, auto

from colorama import Back, Cursor, Fore, Style as AnsiStyle

from . import game

_ROWS = 3
_COLUMNS = 5
_PADDING = 0

class Position(Enum):
    """How to position a tile when printing it."""

    STANDALONE = auto()
    FIRST_PRINT = auto()
    EDIT = auto()

    def begin(self, round_: int, index: int, row: int) -> str:
        """Get the string that begins the positioning."""
        cls = self.__class__

        if self == cls.STANDALONE:
            return ''

        if self == cls.FIRST_PRINT:
            if index == 0:
                if round_ == 0 or row != 0:
                    return ''

                return '\n' * _PADDING

            # pylint: disable=invalid-name
            up = Cursor.UP(_ROWS - row)
            forward = Cursor.FORWARD((_COLUMNS + _PADDING) * index)
            return f'{up}{forward}'

        if self == cls.EDIT:
            # pylint: disable=invalid-name
            up = Cursor.UP(((_ROWS + _PADDING) * (game.ROUNDS - round_)) - _PADDING - row)
            # FIXME: Why do we need the extra FORWARD and BACK?
            forward = Cursor.FORWARD(((_COLUMNS + _PADDING) * index) + 1)
            back = Cursor.BACK()
            return f'{up}{forward}{back}'

        raise ValueError(f"unhandled position: {self.name}")

    def end(self, round_: int, index: int, row: int) -> str:
        """Get the string that ends the positioning."""
        cls = self.__class__

        if self == cls.STANDALONE:
            return '\n'

        if self == cls.FIRST_PRINT:
            if index == 0:
                return '\n'

            down = Cursor.DOWN(_ROWS - row)
            back = Cursor.BACK((_COLUMNS * (index + 1)) + (_PADDING * index))
            return f'{down}{back}'

        if self == cls.EDIT:
            down = Cursor.DOWN(((_ROWS + _PADDING) * (game.ROUNDS - round_)) - _PADDING - row)
            back = Cursor.BACK(((_COLUMNS + _PADDING) * (index + 1)) - _PADDING)
            return f'{down}{back}'

        raise ValueError(f"unhandled position: {self.name}")

class Style(Enum):
    """Style in which to print a tile."""

    INPUT = auto()
    CORRECT = auto()
    MISPLACED = auto()
    ABSENT = auto()

    def __str__(self) -> str:
        cls = self.__class__

        if self == cls.INPUT:
            return Fore.WHITE + Back.BLACK

        if self == cls.CORRECT:
            return Fore.BLACK + Back.GREEN

        if self == cls.MISPLACED:
            return Fore.BLACK + Back.YELLOW

        if self == cls.ABSENT:
            return AnsiStyle.DIM + Fore.WHITE + Back.BLACK

        raise ValueError(f"unhandled style: {self.name}")

@dataclass
class Tile:
    """A printable tile representing a single letter."""

    round_: int
    index: int
    letter: str = ' '
    position: Position = Position.STANDALONE
    style: Style = Style.INPUT

    def __post_init__(self):
        self.letter = self.letter.upper()

    def __str__(self) -> str:
        begin = lambda row: self.position.begin(self.round_, self.index, row)
        end = lambda row: self.position.end(self.round_, self.index, row)

        return (
            f"{begin(0)}{self.style}┌─{'─'        }─┐{AnsiStyle.RESET_ALL}{end(0)}"
            f"{begin(1)}{self.style}│ {self.letter} │{AnsiStyle.RESET_ALL}{end(1)}"
            f"{begin(2)}{self.style}└─{'─'        }─┘{AnsiStyle.RESET_ALL}{end(2)}"
        )



"""Logic for making guesses."""

import string

from ..common import words
from ..common.hint import Hint

LetterScores = dict[str, int]
WordScores = list[LetterScores]

def filter_guesses(hint: Hint) -> set[str]:
    """Get all possible guesses according to the given hint."""
    filtered_guesses: set[str] = set()
    for word in words.possible_words:
        for index, letter in enumerate(word):
            if letter in hint.absent \
                    and word.count(letter) != len(hint.correct.get(letter, set())) \
                                            + len(hint.misplaced.get(letter, set())):
                # This letter is absent in the answer, so this is not a
                # possible guess
                break

            if hint.inverse_correct.get(index, letter) != letter \
                    or letter in hint.inverse_misplaced.get(index, set()):
                # This letter is not at this index in the answer, so this is
                # not a possible guess
                break
        else:
            for letter in hint.misplaced:
                if letter not in word:
                    # This letter exists in the answer, so this is not a
                    # possible guess
                    break
            else:
                filtered_guesses.add(word)

    return filtered_guesses

def score_strings() -> WordScores:
    """Score all possible strings, regardless of whether they are permitted guesses.

    Scores are computed for each character in a string, independently of one
    another. A higher score indicates a better choice of letter.

    Each character's scores are returned in an arbitrary order.
    """
    # Initialize scores to 0
    scores = list(dict.fromkeys(string.ascii_lowercase, 0) for _ in range(words.LENGTH))

    for word in words.possible_words:
        for i in range(words.LENGTH):
            scores[i][word[i]] += 1

    for i in range(words.LENGTH):
        for letter in string.ascii_lowercase:
            # Score each character by its frequency's closeness to 50%. This
            # means the best character is the one that eliminates closest to
            # half the possible words. Note that this is only true per
            # character, so there is correlation between characters that is not
            # considered.
            scores[i][letter] = -abs((len(words.possible_words) // 2) - scores[i][letter])

    return scores

def sort_guesses(guesses: set[str], scores: WordScores) -> list[str]:
    """Sort the given guesses according to the given scores.

    They are returned in nonascending order.
    """
    scored_guesses = {}
    for guess in guesses:
        scored_guesses[guess] = sum(scores[i][guess[i]] for i in range(words.LENGTH))

    sorted_scored_guesses = sorted(scored_guesses.items(), key=lambda item: item[1], reverse=True)
    return [guess for guess, _ in sorted_scored_guesses]



"""Game tracking."""

from datetime import datetime
from string import ascii_lowercase
from time import sleep
from typing import Literal, Union

from colorama import Style as AnsiStyle

from ..backend.hint import hint
from ..common.hint import Hint
from ..common.words import LENGTH, allowed_guesses, possible_words
from ..frontend.tile import Position, Style, Tile
from .player import Bot, User

# The number of rounds in a game
ROUNDS = 6

class Game:
    """Single instance of a game."""

    def __init__(self, answer: str='today\'s', day: Union[Literal['today'], int]='today', bot=False):
        """Create a new game."""
        if answer == 'today\'s':
            if day == 'today':
                day = (datetime.now() - datetime(2021, 6, 19)).days

            answer = possible_words[day % len(possible_words)]

        self.answer = answer
        self.player = Bot() if bot else User()
        self.guesses = []
        self.hints = []
        self.round = 0
        self.index = 0

    def play(self):
        """Play the game from start to finish."""
        self._initialize_board()

        while self.round != ROUNDS and (not self.guesses or self.guesses[-1] != self.answer):
            self.guesses.append('')
            hint_ = Hint() if self.round == 0 else self.hints[-1]
            next_round = self.round + 1

            self.player.start_guess(self.round, hint_)

            while self.round != next_round:
                self._handle_keypress(self.player.output_key(self.index))

            self.player.stop_guess(self.round - 1)

        print()
        self._summarize()
        print()
        self._generate_report()

    def _initialize_board(self):
        """Print the initial, empty board."""
        for round_ in range(ROUNDS):
            for index in range(LENGTH):
                tile = Tile(round_, index, position=Position.FIRST_PRINT)
                print(tile, end="")

        print(end="", flush=True)

    def _handle_keypress(self, key: str):
        """Handle the player's input to the game board."""
        if key in ascii_lowercase:
            if self.index < LENGTH:
                # Add the letter to our guess
                self.guesses[-1] += key
                tile = Tile(self.round, self.index, letter=key, position=Position.EDIT)
                print(tile, end="", flush=True)
                self.index += 1
        elif key == 'backspace':
            if self.index > 0:
                # Delete the current letter from our guess
                self.index -= 1
                self.guesses[-1] = self.guesses[-1][:-1]
                tile = Tile(self.round, self.index, position=Position.EDIT)
                print(tile, end="", flush=True)
        elif key == 'enter':
            guess = self.guesses[-1]
            if guess in allowed_guesses:
                # Submit the completed guess
                hint_ = hint(self.answer, guess)
                self.hints.append(hint_)

                for index, letter in enumerate(guess):
                    if index in hint_.inverse_correct:
                        style = Style.CORRECT
                    elif index in hint_.inverse_misplaced:
                        style = Style.MISPLACED
                    elif letter in hint_.absent:
                        style = Style.ABSENT
                    else:
                        raise KeyError(f"expected \"{letter}\" to be absent")

                    tile = Tile(self.round, index, letter=letter, position=Position.EDIT, style=style)
                    print(tile, end="", flush=True)
                    # "Flip" tiles
                    sleep(0.1)

                self.round += 1
                self.index = 0

    def _summarize(self):
        if self.guesses[-1] == self.answer:
            print(f"{AnsiStyle.BRIGHT}Victory! 🏆{AnsiStyle.RESET_ALL}")
        else:
            print(f"{AnsiStyle.DIM}Defeat 😣{AnsiStyle.RESET_ALL} The word was {AnsiStyle.BRIGHT}{self.answer}{AnsiStyle.RESET_ALL}.")

    def _generate_report(self):
        """Print the same report as Wordle."""
        print(f"Wordle {possible_words.index(self.answer)} {self.round if self.guesses[-1] == self.answer else 'X'}/{ROUNDS}\n")
        for round_ in range(self.round):
            hint_ = self.hints[round_]
            tiles = ["⬛"] * LENGTH

            for correct_index in hint_.inverse_correct:
                tiles[correct_index] = "🟩"

            for misplaced_index in hint_.inverse_misplaced:
                tiles[misplaced_index] = "🟨"

            print("".join(tiles))



"""User interface."""

from typing import Literal, Union

import colorama
from tap import Tap

from .game import Game

class _ArgumentParser(Tap):
    word: str = 'today\'s'  # which word to make the answer
    day: Union[Literal['today'], int] = 'today'  # which day to use for today's word
    bot: bool = False  # have the bot play instead

    def configure(self):
        self.add_argument('-w', '--word')
        self.add_argument('-d', '--day', type=lambda day: day if day == 'today' else int(day))
        self.add_argument('-b', '--bot')

def cli():
    """Command-line interface."""
    args = _ArgumentParser(description="Interact with Wordle.").parse_args()

    colorama.init()
    Game(answer=args.word, day=args.day, bot=args.bot).play()



"""Game player interface and implementations."""

from abc import ABCMeta, abstractmethod
from queue import Queue
from threading import Thread
from time import sleep

from sshkeyboard import listen_keyboard, stop_listening

from ..common.hint import Hint
from ..common.words import LENGTH
from . import game
from .guess import filter_guesses, score_strings, sort_guesses

class Player(metaclass=ABCMeta):
    """A player of the game."""

    @abstractmethod
    def start_guess(self, round_: int, hint: Hint) -> None:
        """Start a new guess.

        :param round_: the round that's starting
        :param hint: the hint resulting from the previous guess
        """

    @abstractmethod
    def output_key(self, index: int) -> str:
        """Output a key for the current guess.

        If it's
        - a letter, then it'll be appended to the guess at the given index.
        - a backspace, then the last letter guessed will be deleted.
        - an enter, then the current guess will be submitted.
        """

    @abstractmethod
    def stop_guess(self, round_: int) -> None:
        """Conclude the current guess; it has been submitted.

        :param round_: the round that finished
        """

class User(Player):
    """A human—or chimpanzee… maybe even a corvid?—player."""

    def __init__(self):
        self._thread: Thread = None
        self._queue: Queue[str] = Queue()

    def start_guess(self, round_: int, hint: Hint) -> None:
        if round_ == 0:
            self._start_listening()

    def output_key(self, index: int) -> str:
        return self._queue.get()

    def stop_guess(self, round_: int) -> None:
        if round_ == game.ROUNDS - 1:
            self._stop_listening()

    def _start_listening(self) -> None:
        """Start listening to the user's keyboard in a new thread."""
        self._thread = Thread(
            target=listen_keyboard,
            kwargs={
                'on_press': self._handle_keypress,
                'delay_second_char': 0.05,
            },
            daemon=True,
        )
        self._thread.start()

    def _stop_listening(self) -> None:
        """Stop listening to the user's keyboard."""
        stop_listening()

    def _handle_keypress(self, key: str) -> None:
        """Handle the user's input."""
        self._queue.put(key)

class Bot(Player):
    """A bot: No I/O is required; it guesses on its own."""

    def __init__(self):
        self._possible_guesses: set[str] = filter_guesses(Hint())
        self._guess: str = None

    def start_guess(self, round_: int, hint: Hint) -> None:
        self._possible_guesses &= filter_guesses(hint)
        sorted_guesses = sort_guesses(self._possible_guesses, score_strings())

        self._guess = sorted_guesses[0]
        self._possible_guesses.remove(self._guess)
        # Pause for effect
        sleep(1)

    def output_key(self, index: int) -> str:
        # Pause for effect
        sleep(0.25)
        return 'enter' if index == LENGTH else self._guess[index]

    def stop_guess(self, round_: int) -> None:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['wyrdle', 'wyrdle.backend', 'wyrdle.common', 'wyrdle.frontend']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.4,<0.5.0',
 'sshkeyboard>=2.3.1,<3.0.0',
 'typed-argument-parser>=1.7.2,<2.0.0']

entry_points = \
{'console_scripts': ['wordle = wyrdle.frontend.interface:cli',
                     'wyrdle = wyrdle.frontend.interface:cli']}

setup_kwargs = {
    'name': 'wyrdle',
    'version': '1.2.0',
    'description': 'Wordle and bot in Python',
    'long_description': "# Wyrdle\n![Funny badge](https://forthebadge.com/images/badges/60-percent-of-the-time-works-every-time.svg)\n\nThis is a Python implementation of the viral word game, [Wordle](https://www.powerlanguage.co.uk/wordle/), by [Josh Wardle](https://www.powerlanguage.co.uk/) (ohh, so that's where the name is from).\n\nAdditionally, a bot is included that tries to win the game with its own strategy!\n\n## Installation\nInstall from PyPI using\n```sh\npip install wyrdle\n```\n\n## Usage\nYou can play today's Wordle right from your terminal! Simply run\n```sh\nwordle\n```\n\nFeeling lazy? Feeling competitive? (We're going full Google search button here.)  \nThen let the bot play:\n```sh\nwordle -b\n```\n\n### Example\nHere is the bot's run against `smile`:\n> Wordle 666 2/6  \n>   \n> 🟩🟨⬛⬛🟩  \n> 🟩🟩🟩🟩🟩  \n\nImpressive!\n",
    'author': 'Cole French',
    'author_email': '16979554+ColeFrench@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ColeFrench/wyrdle',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
